////////////////////////////////////////////////////////////////////////////
//
// FILE NAME
//    Collimator.cc
//
// AUTHOR
//    Sarah Cousineau, ORNL, scousine@sns.gov
//    Jeff Holmes, ORNL, jzh@sns.gov
//
// CREATED
//    11/1/00
//
//  DESCRIPTION
//    Finite length collimation model.  The structure of this code is
//    modeled after the K2 collimation code written by Jean-Bernard 
//    Jeanneret at CERN.  A description of the K2 code can be found 
//    in the PhD thesis of Nuria Catalan-Lasheras.  Much of the 
//    information can be found also in the Particle Data Book.  Models for 
//    nuclear scattering have been adapted specifically for low energy,
//    with low energy elastic scattering modeled after the LAHET code,
//    and cross sections for the nuclear data generated by MCNPX.    
//    A more complete description of the routines and list of references 
//    will soon be available in the ORBIT user's guide.
//
//  REVISION HISTORY
// 
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// Include files
//
/////////////////////////////////////////////////////////////////////////////
#include "Injection.h"
#include "Collimator.h"
#include "Aperture.h"
#include "Node.h"
#include "MacroPart.h"
#include "RealMat.h"
#include "StreamType.h"
#include "MapBase.h"
#include "TransMapHead.h"
#include "TeaPotHead.h"
#include "Spline.h"
#include <iostream>
#include <iomanip>
#include <cmath>
#include <ctime>

//MPI header file for parallel stuff ==MPI==
#include "mpi.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////
//
//
// STATIC DEFINITIONS
//
///////////////////////////////////////////////////////////////////////////

Array(ObjectPtr) CollimatorPointers;
extern Array(ObjectPtr) syncP, mParts, nodes, tMaps, AperturePointers;

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
#define JMAX 40
#define PI 3.1415926539

///////////////////////////////////////////////////////////////////////////
//
// Local routines:
//
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//
// CLASS NAME
//    CollimatorBase 
//
// INHERITANCE RELATIONSHIPS
//    CollimatorBase  -> Node* -> Object
//
// USING/CONTAINING RELATIONSHIPS
//    None.
//
// DESCRIPTION
//    This is a base class for storing Collimator information. 
//
// PUBLIC MEMBERS
//
// PROTECTED MEMBERS
//    _nHits          - number of hits.
//    _nLost          - number of particles lost.
//    _showCollimator - "Routine to print out the aperture parameters"
//    
// PRIVATE MEMBERS
//    None.
//
// 
///////////////////////////////////////////////////////////////////////////

class CollimatorBase : public Node {
  Declare_Standard_Members(CollimatorBase, Node);
public:
    CollimatorBase(const String &n, const Integer &order, 
		   const Real &l):
            Node(n, l, order)
            
        {
	  _nHits=0;
	  _nLost=0;
          _nHitsGlobal = 0;
          _nLostGlobal = 0;
        }

    virtual Void nameOut(String &wname) { wname = _name; }
    virtual Void _updatePartAtNode(MacroPart &mp)=0;
    virtual Void _nodeCalculator(MacroPart &mp)=0;    
    virtual Void _showCollimator(Ostream &sout)=0;
    static Void _syncLostOrHits( Integer &nLoc, Integer &nGlob);
    Integer _nHits;  
    Integer _nLost;
    Integer _nHitsGlobal;
    Integer _nLostGlobal;
    
};

///////////////////////////////////////////////////////////////////////////
//
// NON-INLINE PROTECTED MEMBER FUNCTIONS FOR CLASS CollimatorBase
//
///////////////////////////////////////////////////////////////////////////

Define_Standard_Members(CollimatorBase, Node);

Void CollimatorBase::_syncLostOrHits( Integer &nLoc, Integer &nGlob)
{
  nGlob = nLoc;
  if( Collimator::nMPIsize > 1 ){
          MPI_Allreduce(&nLoc,&nGlob,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
  }
}

///////////////////////////////////////////////////////////////////////////
//
// CLASS NAME
//    Collimate 
//
// INHERITANCE RELATIONSHIPS
//    Collimate  -> CollimatorBase -> Node* -> Object
//
// USING/CONTAINING RELATIONSHIPS
//    None.
//
// DESCRIPTION
//    This is a class for collimator information. 
//
// PUBLIC MEMBERS
//    _radius:     radius of a circular collimator [mm]
//    _ma:         material number
//    _densityfac: equivalent density of material
//    _shape:      collimator shape
//    _a:          bore width, depends on shape
//    _b:          bore width, depends on shape
//    _c:          bore width, depends on shape
//    _d:          bore width, depends on shape
//    _angle:      rotation angle
//
// PROTECTED MEMBERS
//  None
// PRIVATE MEMBERS
//  None.
//
///////////////////////////////////////////////////////////////////////////

class Collimate : public CollimatorBase {
  Declare_Standard_Members(Collimate, CollimatorBase);
public:
  Collimate(const String &n, const Integer &order,
	    const Real &l, const Integer &ma, 
	    const Real &densityfac, const Integer &shape, 
	    const Real &a, const Real &b, const Real &c,
	    const Real &d, const Real &angle):
    CollimatorBase(n, order, l), _ma(ma), _densityfac(densityfac), 
    _shape(shape), _a(a), _b(b), _c(c), _d(d), _angle(angle)
    
  {
    crossEn.resize(57);
    crossEl.resize(57);
    crossInel.resize(57);
    crossRuth.resize(57);
    _nHits=0;
    _getMaterialParams(_ma);
  }

  virtual Void _updatePartAtNode(MacroPart &mp);
  virtual Void _nodeCalculator(MacroPart &mp);
  virtual Void _showCollimator(Ostream &sout);
  virtual Void _getMaterialParams(Integer &);
  
  Integer _ma, _shape;
  Real _a, _b, _c, _d, _angle, _A, _Z, _R;
  Real _density, _densityfac, _radlength;
  Vector(Real) crossEn, crossEl, crossInel, crossRuth;
};

///////////////////////////////////////////////////////////////////////////
// 
// NON-INLINE PROTECTED MEMBER FUNCTIONS FOR CLASS Collimate 
//
///////////////////////////////////////////////////////////////////////////

Define_Standard_Members(Collimate, Node);

///////////////////////////////////////////////////////////////////////////
// 
// NAME
//    Collimate::NodeCalculator
//
// DESCRIPTION
//    Empty
//
// PARAMETERS
//    None
// 
// RETURNS
//    Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimate::_nodeCalculator(MacroPart &mp) { }

///////////////////////////////////////////////////////////////////////////
//
// NAME
//    Collimate::updatePartAtNode
//
// DESCRIPTION
//    Transfer particles through collimator.
//
// PARAMETERS
//    None.
//
// RETURNS
//    Nothing.
//
///////////////////////////////////////////////////////////////////////////

Void Collimate::_updatePartAtNode(MacroPart& mp)
{
  const Real nAvogadro = 6.022045e23;
  Integer j = 1, idum, coll_flag = 0, lastArg, trackit;
  Real eCross, iCross, rCross, e_frac, i_frac, r_frac, b_pN;
  Real totCross, random, choice, length, dlength, meanFreePath;
  Real rl, zrl, stepsize, smallstep, radlengthfac;
  Real p, E, beta, dE, M_nuc, M;
  Real theta, theta_norm, delta_pos, delta_mom, thx, thy;
  Real t, dp_x=0., dp_y=0, pfac, xpfac, ypfac, directionfac;

  Spline ElValue(crossEn, crossEl);
  Spline InelValue(crossEn, crossInel);
  Spline RuthValue(crossEn, crossRuth);

  idum = -(unsigned)time(0);
  M = mp._syncPart._e0;
  M_nuc = _A * 0.931494;
  E = mp._syncPart._eTotal;
  length = _length * 1000.;
  dlength = length * 1.0e-04;
  radlengthfac = _radlength * 10. / _densityfac;
  smallstep = 0.001 * radlengthfac;
  if(smallstep > length / 10.) smallstep = length / 10.;
  if(_A <= 62.) b_pN = 14.5 * pow(_A, 0.6666667);
  if(_A >  62.) b_pN = 60.0 * pow(_A, 0.3333333);

  while(j <= mp._nMacros)
  {
    zrl = length;
    coll_flag = Collimator::checkCollFlag(_shape, _a, _b, _c, _d,
					  _angle, mp._x(j), mp._y(j));
    if(coll_flag == 1)
    {
      _nHits++;
      if((_ma == 9) && (!TransMap::tuneCalcOn))
      {
	mp._addLostMacro(j, _position);
	j--;
	_nLost++;
        zrl = -1.;
      }
    }

    while(zrl > 0)
    {
      //Drift.
      while((coll_flag == 0) && (zrl > 0))
      {
	stepsize = 1.;
        if(stepsize > length / 10.) stepsize = length / 10.;
        if(stepsize > zrl) stepsize = zrl + dlength;
        pfac = 1.0 + mp._dp_p(j);
	mp._x(j) += stepsize * mp._xp(j) / (1000.* pfac);
	mp._y(j) += stepsize * mp._yp(j) / (1000.* pfac);
	zrl -= stepsize;
        coll_flag = Collimator::checkCollFlag(_shape, _a, _b, _c,_d, 
					      _angle, mp._x(j), mp._y(j));
        if(coll_flag == 1)
	{
	  _nHits++;
	  if((_ma == 9) && (!TransMap::tuneCalcOn))
          {
            mp._addLostMacro(j, _position);
            j--;
            _nLost++;
	    zrl = -1.;
          }
        }
      }

      //Collimate
      while((coll_flag == 1) && (zrl > 0))
      {
        E = mp._syncPart._eTotal + mp._deltaE(j);
        p = pow((E * E - M * M), 0.5);
        beta = p / E;
        pfac = 1.0 + mp._dp_p(j);
        xpfac = mp._xp(j) / (1000.* pfac);
        ypfac = mp._yp(j) / (1000.* pfac);
        directionfac = sqrt(1.0 + xpfac * xpfac + ypfac * ypfac);
        rl = zrl * directionfac;

        theta = 0.0136 / (beta * p)
                / sqrt(radlengthfac);
        eCross = ElValue(E - M);
        iCross = InelValue(E - M);
	/*
        theta_norm = theta * sqrt(radlengthfac);
        //lastArg = 0;
	lastArg=2;
        rCross = Collimator::getRuth_t(theta_norm, E, p, _R, _Z,
                                       idum, lastArg);
	//rCross = RuthValue(E - M);
	*/
        stepsize = rl;
        trackit = 0;
        Collimator::getRuthJackson(stepsize, _Z, _A, _density,
		                   idum, beta, trackit,
                                   rCross, thx, thy);
        totCross = eCross + iCross + rCross;
        meanFreePath = (_A / nAvogadro /
                       (_density * _densityfac * 1.0e+06) /
                       (totCross * 1.0e-28)) * 1000.;
        stepsize = -meanFreePath * log(Collimator::ran1(idum));

        if(stepsize > rl)
        {
          // MCS out, no nuclear interaction.
          while((coll_flag == 1) && (zrl > 0))
          {
            stepsize = rl + dlength;
            E = mp._syncPart._eTotal + mp._deltaE(j);
            p = pow((E * E - M * M), 0.5);
            beta = p / E;
            theta = 0.0136 / (beta * p)
                    / sqrt(radlengthfac);

            Collimator::checkStep(rl, stepsize, length, _shape, 
                                  _a, _b, _c, _d, _angle, mp._x(j), 
                                  mp._y(j), mp._xp(j), mp._yp(j), 
                                  theta);

            if(stepsize < smallstep) stepsize = smallstep;
            if(stepsize > rl) stepsize = rl + dlength;
	    if(stepsize > 0.0)
            {
              pfac = 1.0 + mp._dp_p(j);

              // MCS
	      /*
              Collimator::MCS(stepsize, theta, idum, mp._xp(j),
                              delta_pos, delta_mom);
              mp._x(j)  += delta_pos;
              mp._xp(j) += delta_mom * pfac;
              Collimator::MCS(stepsize, theta, idum, mp._yp(j),
                              delta_pos, delta_mom);
	      mp._y(j)  += delta_pos;
              mp._yp(j) += delta_mom * pfac;
	      */
              Collimator::MCSJackson(stepsize, _Z, _A, _density,
                                     idum, beta, pfac,
                                     mp._x(j), mp._y(j),
                                     mp._xp(j), mp._yp(j));

	      //Ionization energy loss
              dE = Collimator::IonEnergyLoss(beta, _Z, _A);
              dE = -dE * _density * _densityfac / 10000. * stepsize;
              mp._deltaE(j) += dE;
              mp._dp_p(j) = mp._deltaE(j) * mp._syncPart._dppFac;

              // Remove particle below 20 MeV
              if(((mp._deltaE(j) + mp._syncPart._eKinetic) < 0.02) &&
		 (!TransMap::tuneCalcOn))
              {
                mp._addLostMacro(j, _position);
                j--;
                _nLost++;
                coll_flag == 0;
                zrl = -1.;
              }
              if((coll_flag == 1) && (zrl > 0))
              { 
                coll_flag = Collimator::
                            checkCollFlag(_shape, _a, _b, _c, _d, _angle,
                                          mp._x(j), mp._y(j));
                E = mp._syncPart._eTotal + mp._deltaE(j);
                p = pow((E * E - M * M), 0.5);
                pfac = 1.0 + mp._dp_p(j);
                xpfac = mp._xp(j) / (1000. * pfac);
                ypfac = mp._yp(j) / (1000. * pfac);
                directionfac = sqrt(1.0 + xpfac * xpfac + ypfac * ypfac);
                zrl -= stepsize / directionfac;
                rl = zrl * directionfac;
              }
            }
            else
            {
              coll_flag = 0;
              zrl = -1;
            }
          }
        }

        else

        {
          // In collimator, allow nuclear interactions
          E = mp._syncPart._eTotal + mp._deltaE(j);
          p = pow((E * E - M * M), 0.5);
          beta = p / E;
          theta = 0.0136 / (beta * p)
                  / sqrt(radlengthfac);

          pfac = 1.0 + mp._dp_p(j);

          // MCS
	  /*
          Collimator::MCS(stepsize, theta, idum, mp._xp(j),
                          delta_pos, delta_mom);
          mp._x(j)  += delta_pos;
          mp._xp(j) += delta_mom * pfac;
          Collimator::MCS(stepsize, theta, idum, mp._yp(j),
                          delta_pos, delta_mom);
          mp._y(j)  += delta_pos;
          mp._yp(j) += delta_mom * pfac;
	  */
          Collimator::MCSJackson(stepsize, _Z, _A, _density,
                                 idum, beta, pfac,
                                 mp._x(j), mp._y(j),
                                 mp._xp(j), mp._yp(j));

          //Ionization energy loss
          dE = Collimator::IonEnergyLoss(beta, _Z, _A);
          dE = -dE * _density * _densityfac / 10000. * stepsize;
          mp._deltaE(j) += dE;
          mp._dp_p(j) = mp._deltaE(j) * mp._syncPart._dppFac;

          // Remove particle below 20 MeV
          if(((mp._deltaE(j) + mp._syncPart._eKinetic) < 0.02) &&
	    (!TransMap::tuneCalcOn))
          {
            mp._addLostMacro(j, _position);
            j--;
            _nLost++;
            coll_flag == 0;
            zrl = -1.;
          }
          else
          { 
            E = mp._syncPart._eTotal + mp._deltaE(j);
            p = pow((E * E - M * M), 0.5);
            beta = p / E;
            theta = 0.0136 / (beta * p)
                    / sqrt(radlengthfac);

            pfac = 1.0 + mp._dp_p(j);
            xpfac = mp._xp(j) / (1000. * pfac);
            ypfac = mp._yp(j) / (1000. * pfac);
            directionfac = sqrt(1.0 + xpfac * xpfac + ypfac * ypfac);
            zrl -= stepsize / directionfac;
            rl = zrl * directionfac;
            coll_flag = Collimator::checkCollFlag(_shape, _a, _b,
                                                  _c, _d, _angle,
                                                  mp._x(j), mp._y(j));
          }

          //Nuclear interact
          if(coll_flag==1 && zrl > 0)
          {
            eCross = ElValue(E - M);
            iCross = InelValue(E - M);
	    /*
            theta_norm = theta * sqrt(radlengthfac);
            //lastArg = 0;
	    lastArg = 2;
            rCross = Collimator::getRuth_t(theta_norm, E, p, _R, _Z,
                                           idum, lastArg);
	    //rCross = RuthValue(E - M);
	    */
            trackit = 0;
            Collimator::getRuthJackson(stepsize, _Z, _A, _density,
	                               idum, beta, trackit,
                                       rCross, thx, thy);
            totCross = eCross + iCross + rCross;
            e_frac = eCross/totCross;
            i_frac = iCross/totCross;
            r_frac = rCross/totCross;
	    choice = Collimator::ran1(idum);

            pfac = 1.0 + mp._dp_p(j);

	    // Nuclear Elastic Scattering
            if((choice >= 0.) && (choice <= e_frac))
            {
              if((E - M) <= 0.4)
              {
                t=Collimator::getElast_t(p, _A, idum);
              }
              if((E - M) > 0.4)
              {
                t=-log(Collimator::ran1(idum))/b_pN;
              }
              Collimator::getKick(t, p, M_nuc, dp_x, dp_y, idum);
              mp._xp(j) += dp_x * pfac;
              mp._yp(j) += dp_y * pfac;
            }

            // Rutherford Coulomb scattering
            if((choice > e_frac) && (choice <= (1 - i_frac)))
            {
	      /*
              theta_norm = theta * sqrt(radlengthfac);
              lastArg = 1;
	      //lastArg = 3;
	      t = Collimator::getRuth_t(theta_norm, E, p, _R, _Z,
                                        idum, lastArg);
              Collimator::getKick(t, p, M_nuc, dp_x, dp_y, idum);
              mp._xp(j) += dp_x * pfac;
              mp._yp(j) += dp_y * pfac;
	      */
              trackit = 1;
              Collimator::getRuthJackson(stepsize, _Z, _A, _density,
	                                 idum, beta, trackit,
                                         rCross, thx, thy);

              xpfac = mp._xp(j) / (1000.* pfac);
              ypfac = mp._yp(j) / (1000.* pfac);

              Real anglex = atan(xpfac) + thx;
              Real angley = atan(ypfac) + thy;

              mp._xp(j) = tan(anglex) * (1000.* pfac);
              mp._yp(j) = tan(angley) * (1000.* pfac);
            }

            // Nuclear Inelastic absorption
            if((choice > (1.-i_frac)) && (choice <= 1.)
			 && !TransMap::tuneCalcOn)
            {
              mp._addLostMacro(j, _position);
              j--;
              _nLost++;
              zrl = -1.;
            }
          }
        }
      }
    }
    j++;
  }
}

///////////////////////////////////////////////////////////////////////////
//
// NAME
//    Collimate::_showCollimator
//
// DESCRIPTION
//    Prints information about collimators.
//
// PARAMETERS
//    None.
//
// RETURNS
//    Nothing.
//
///////////////////////////////////////////////////////////////////////////

Void Collimate::_showCollimator(Ostream &sout)
{    

  _syncLostOrHits(_nHits,_nHitsGlobal);

  if(Collimator::nMPIrank == 0) {
    sout << " Collimator:"  << "\n";
    sout << " Position in ring [m]=" << _position << "\n";
    sout << " Length of Collimator [m] =" << _length <<"\n";
    sout << " Number of macroparticle hits = " << _nHitsGlobal << "\n";
  }
}


///////////////////////////////////////////////////////////////////////////
//
// PUBLIC MEMBER FUNCTIONS FOR MODULE Collimator
//
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//
// NAME
//    Collimator::ctor
//
// DESCRIPTION
//    Initializes the various Collimator related constants.
//
// PARAMETERS
//    None.
//
// RETURNS
//    Nothing.
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::ctor()
{
// set some initial values

    nCollimators = 0;

  // parallel stuff  ==MPI==  ==start===
    iMPIini=0;
    MPI_Initialized(&iMPIini);
    nMPIsize = 1;
    nMPIrank = 0;
    if(iMPIini > 0){
      MPI_Comm_size(MPI_COMM_WORLD,&nMPIsize);
      MPI_Comm_rank(MPI_COMM_WORLD,&nMPIrank);
    }
  // parallel stuff  ==MPI==  ==stop===
    
}

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//  Collimator::addCollimator
//
// DESCRIPTION
//    Adds a collimator.
//
// PARAMETERS
//    name:       Name for this node (string)
//    order:      node order index (integer)
//    l:          length of collimator (real)
//    ma: 	  material number (integer)
//    densityfac: equivalent density (real)
//    shape     : collimator transverse shape (integer)
//    a         : shape parameter 1.
//    b         : shape paremeter 2.
//    c         : shape parameter 3.
//    d         : shape parameter 4.
//    angle     : transverse shape title angle.
//
// RETURNS
//    Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::addCollimator(const String &name,
			       const Integer &order, const  Real &l,
			       const Integer &ma, const Real &densityfac,
			       const Integer &shape, const Real &a,
			       const Real &b, const Real &c,
			       const Real &d, const Real &angle)
{
  //Set the nearest transfer matrix to a unitary matrix.

  Integer i, j, k;
  MapBase *tm;

  Integer transmapNumber = -1;

  for(i = 1; i <= nTransMaps; i++)
  {
    tm = MapBase::safeCast(tMaps(i - 1));
    if(tm->_oindex == order) transmapNumber = i;
  }

  if(transmapNumber < 0) except(badTransMap);
  if(transmapNumber > nTransMaps) except(badTransMap);

  tm = MapBase::safeCast(tMaps(transmapNumber - 1));
  tm -> _length = -Consts::tiny;

   //Add the collimator node.

   if(nNodes == nodes.size()) nodes.resize(nNodes + 1);
   if(nCollimators == CollimatorPointers.size())
     CollimatorPointers.resize(nCollimators + 1);
   nNodes++;

   nodes(nNodes - 1) = new Collimate(name, order, l, ma, densityfac, shape,
				   a, b, c, d, angle);
   nCollimators++;

   CollimatorPointers(nCollimators - 1) = nodes(nNodes - 1);

   nodesInitialized = 0;
}

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::checkCollFlag
//
// DESCRIPTION
//   Checks to see if a particle is located inside a collimator.  Returns
//   1 if the particle is in the collimator, 0 if it isn't.
//
// PARAMETERS
//   shape:  shape of the collimator: 1=circle, 2=ellipse, 3=one sided
//           flat, 4=two sided flat, 5=rectangular (outside is collimator),
//           6=rectangular (inside is collimator).
//   a:      depending on shape, either (shape = 1) radius, 
//           (shape = 2) semimajor axis, (shape = 3) distance to 
//           flat edge, (shape = 4) minimum edge, (shape=5 or 6) 
//           minimum horizontal edge.
//   b:      depending on shape, either (1) radius, (2) semimajor axis,
//           (3) zero  (4) maximum edge (5) (shape=5 or 6) maximum 
//           horizontal edge.
//   c:      minimum vertical edge (used only in shapes 5 or 6)
//   d:      maximum vertical edge (used only in shapes 5 or 6)
//
//   angle:  tilt angle of collimator.
//   x:      x coordinate of particle.
//   y:      y coordinate of particle.
//
// RETURNS
//   Integer.
//
///////////////////////////////////////////////////////////////////////////

Integer Collimator::checkCollFlag(const Integer &shape, const Real &a, 
				  const Real &b, const Real &c, 
				  const Real &d, const Real &angle, 
				  Real &x, Real &y)
{ 
  Real xtemp=x, ytemp=y;
 
  if(shape==1)
    {
      if( (Sqr(x)+Sqr(y)) >= Sqr(a)){
	return 1;
      }
      else 
	return 0;
    }
  if(shape==2)
    {
      if(angle != 0)
	{
	  xtemp =  cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      if( (Sqr(xtemp/a)+Sqr(ytemp/b)) >= 1.)
      	return 1;
      else return 0;
    }
  if(shape==3)
    {
      if(angle != 0)
	{
	  xtemp =  cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      if( xtemp >= a) 
	return 1;
      else return 0;
    }
  if(shape==4)
    {
      if(angle != 0)
	{
	  xtemp =  cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      if( xtemp <= a || xtemp >= b) 
	return 1;
      else return 0;
    }
 
  if(shape==5)
    {
      if(angle != 0)
	{
	  xtemp =  cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      if( xtemp <=  a ||
          xtemp >=  b ||
          ytemp <=  c ||
          ytemp >=  d )       
	return 1;
      else return 0;
    }

  if(shape==6)
    {
      if(angle != 0)
	{
	  xtemp =  cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      //cout<<"xtemp is, "<<xtemp<<", ytemp is, "<<ytemp<<"\n";
      if( xtemp >=  a &&
          xtemp <=  b &&
          ytemp >=  c &&
          ytemp <=  d )       
	return 1;
      else return 0; 
    }
}  


///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::checkStep
//
// DESCRIPTION
//   Checks to make sure that the randomly generated step size isn't too
//   large (See Catalan-Lasheras, PhD thesis, p 52).  The model is based
//   on the K2 collimation code written by Jean Bernard Jeanneret.
//
// PARAMETERS
//   
//   s:        Length of collimator transversed (s=0 at start of collimator)
//   stepsize: Randomly generated step
//   length:   Collimator length
//   shape:  shape of the collimator: 1=circle, 2=ellipse, 3=one sided
//           flat, 4=two sided flat.
//   a:      depending on shape, either (shape = 1) radius, 
//           (shape = 2) semimajor axis, (shape = 3) distance to edge, 
//           (shape = 4) minimum edge, (shape=5 or 6) minimum 
//           horizontal edge.
//   b:      depending on shape, either (1) nothing (2) semimajor axis,
//           (3) nothing (4) maximum edge, (shape=5 or 6) maximum horizontal 
//           edge.           
//   c:      minimum horizontal edge (used only in shapes 5 or 6)
//   d:      maximum vertical edge (used only in shapes 5 or 6)
//   
//   angle:  tilt angle of collimator.
//   x:      x coordinate of particle.
//   y:      y coordinate of particle.
//   px:     x momentum coordinate.
//   py:     y momentum coordinate.
//   theta:    Related to MCS.  (See Catalan-Lasheras thesis, p52.)
//
// RETURNS
//   Integer.
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::checkStep(Real &s, Real &stepsize, const Real &length, 
			   const Integer &shape, const Real &a, const
			   Real &b, const Real &c, const Real &d,
			   const Real &angle, Real &x, Real &y, 
			   Real &px, Real &py, Real &theta)
{ 
     
  float fstep(float, float, float, float);
  int zbrak(float (*)(float, float, float, float), float, float, 
	    int, float *, float *, int &, float, float, float);
  float rtbis(float (*)(float, float, float, float), float, float, 
	      float xacc, float, float, float);
  float x1=0;
  float x2=length - s;
  int n=50;
  int nb=0;
  float xb1[4];
  float xb2[4];
  float solution[4];
  float smax=0.;
  float xacc=.001;
  int i;
  float r_o, pr_o;
  Real arg, theta_o,  x_ellipse, y_ellipse;
  Real xtemp=x, ytemp=y;

  
  if(shape==1)
    {
      r_o = sqrt(x*x + y*y) - a;
      if(a==0.) r_o=10000000.;
    }


  if(shape==2)
    {
      if(angle != 0)
	{
	  xtemp = cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      arg = ytemp/xtemp;
      if(xtemp>=0) theta_o = atan(arg);
      if(xtemp<0) theta_o = PI+atan(arg);
      y_ellipse = sqrt(Sqr(a*tan(theta_o))/(1+Sqr(a*tan(theta_o)/b)));
      x_ellipse = sqrt(a*a - Sqr(a*y_ellipse/b));
      r_o = sqrt(xtemp*xtemp + ytemp*ytemp) - 
	sqrt(x_ellipse*x_ellipse + y_ellipse*y_ellipse);
    }	       


  if(shape==3)
    {
      if(angle != 0)
	{
	  xtemp = cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      r_o = fabs(xtemp - a);
    }


  if(shape==4)
    {
      if(angle != 0)
	{
	  xtemp = cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      if(xtemp<=a) r_o = fabs(xtemp - a);
      if(xtemp>=b) r_o = fabs(xtemp - b);
    }


 if(shape==5)
    {
      if(angle != 0)
	{
	  xtemp = cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}

      if(xtemp<=a || xtemp>=b){
	if(xtemp<=a){
	  if(ytemp<=c) 
	    r_o=sqrt(Sqr(xtemp - a) + Sqr(ytemp - c));
	  if(ytemp>=d)
	    r_o=sqrt(Sqr(xtemp - a) + Sqr(ytemp - d));
	  if(ytemp>=c && ytemp<=d)
	    r_o = fabs(xtemp - a);
	}
	if(xtemp>=b){
	  if(ytemp<=c)
	    r_o=sqrt(Sqr(xtemp - b) + Sqr(ytemp - c));
	  if(ytemp>=d)
	    r_o=sqrt(Sqr(xtemp - b) + Sqr(ytemp - d));
	  if(ytemp>=c && ytemp<=d)
	    r_o = fabs(xtemp - b);
	}
      }
      else{
	if(ytemp<=c) r_o = fabs(ytemp - c);
	if(ytemp>=d) r_o = fabs(ytemp - d);
      }
    }


  if(shape==6)
    {
      if(angle != 0)
	{
	  xtemp = cos(angle*PI/180.)*x + sin(angle*PI/180.)*y;
	  ytemp = -sin(angle*PI/180.)*x + cos(angle*PI/180.)*y;
	}
      r_o=fabs(xtemp - a);
      if(r_o > fabs(xtemp - b)) r_o = fabs(xtemp - b);
      if(r_o > fabs(ytemp - c)) r_o = fabs(ytemp - c);
      if(r_o > fabs(ytemp - d)) r_o = fabs(ytemp - d);
    }

  pr_o = (x*px/1000. + y*py/1000.)/sqrt(x*x+y*y);

  for(i=0; i<4; i++)
    {
      xb1[i]=0;
      xb2[i]=0;
    }
   
  zbrak(fstep, x1, x2, n, xb1, xb2, nb, r_o, pr_o, theta);

  for(i=1; i<=nb; i++)
    {
      solution[i]=rtbis(fstep, xb1[i], xb2[i], xacc, r_o, pr_o, theta);
    }
  
  solution[0]=100000.;

  for(i=1; i<=nb; i++)
    if(solution[i] < solution[i-1])
      {
	smax=solution[i];
      }
 
  if(smax > 0 && stepsize > smax) stepsize=smax;
 
 
}

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::MCS
//
// DESCRIPTION
//   Multiple coulomb scatter a particle through a set distance.  See
//   the Particle Data Book or the description of the K2 code in 
//   the PhD thesis of Nuria Catalan-Lasheras.
//
// PARAMETERS
//   stepsize: length of material to scatter through.
//   xx:  position coordinate
//   xp:  momentum coordinate
//   theta: angle used in coulomb scattering (See Catalan-Lasheras)
//
// RETURNS
//   Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::MCS(Real &stepsize, Real &theta, Integer &idum, Real &p_o,
		     Real &delta_pos, Real &delta_mom)
{ 
  real v1, v2, z1, z2, rsqrd;
 
  rsqrd=2.;
  while(rsqrd > 1)                            //Generate random, gaussian
    {                                         //distributed number.
      v1 = 2.*Collimator::ran1(idum) - 1;
      v2 = 2.*Collimator::ran1(idum) - 1;  
      rsqrd = v1*v1 + v2*v2;              
    }
  z1 = v1*sqrt(-2.*log(rsqrd)/rsqrd); 
  z2 = v2*sqrt(-2.*log(rsqrd)/rsqrd);
  delta_pos= stepsize*p_o/1000. 
    + theta*stepsize*sqrt(stepsize)/2.*(z1/sqrt(3.)+z2);
  delta_mom = theta*sqrt(stepsize)*z2*1000;
}  

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::MCSJackson
//
// DESCRIPTION
//   Multiple coulomb scatter a particle through a set distance.
//   Simple formulation from JD Jackson:
//   Classical Electrodynamics, Chapter 13.
//   Units are cgs, esu.
//
// PARAMETERS
//   stepsize: length (thickness) of scattering material {mm}.
//   Z:        atomic number of scattering material.
//   A:        atomic weight of scattering material.
//   rho:      mass density of scattering material.
//   idum:     random number seed.
//   beta:     v/c.
//   pfac:     1+dp/p0.
//   x,y:      horizontal and vertical coordinates {mm}.
//   px,py:    horizontal and vertical momenta {mradian}.
//
// RETURNS
//   Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::MCSJackson(Real &stepsize, Real &Z, Real &A, Real &rho,
                            Integer &idum, Real &beta, Real &pfac,
                            Real &x, Real &y, Real &px, Real &py)
{
  Real pi = Consts::pi;
  Real hbar = 1.05443e-27;
  Real cvel = Consts::vLight * 100.0;
  Real eesu = 4.80286e-10;
  Real AMU = 1.65979e-24;
  Real aBohr = 5.29172e-09;
  Real AP = 1.007593;
  Real gamma = 1.0 / sqrt(1.0 - beta * beta);

  Real aMax = 1.4 * aBohr / pow(Z, 0.333333);
  Real aMin = 1.4e-13 * pow(A, 0.333333);

  Real N = rho / (A * AMU);
  Real T = stepsize / 10.0;

  Real pmom = gamma * beta * cvel * AP * AMU;

  Real thMin = hbar / (pmom * aMax);
  Real thMax = hbar / (pmom * aMin);

  Real thMin2i = 1.0 / (thMin * thMin);
  Real thMax2i = 1.0 / (thMax * thMax);
  Real th2iDiff = thMin2i - thMax2i;

  Real th2s = 2.0 * log(thMax / thMin) / th2iDiff;
  Real coeff = 2.0 * Z * eesu * eesu / (pmom * beta * cvel);

  Real sigTot = pi * coeff * coeff * th2iDiff;
  Real nColl = N * T * sigTot;

  Real th2Tot = nColl * th2s;

  Real probrp = Collimator::ran1(idum);
  Real probxy = 2.0 * pi * Collimator::ran1(idum);

  Real angle = sqrt(-th2Tot * log(probrp));
  Real anglexMCS = angle * cos(probxy);
  Real angleyMCS = angle * sin(probxy);

  Real xpfac = px / (1000.* pfac);
  Real ypfac = py / (1000.* pfac);

  Real anglex = atan(xpfac) + anglexMCS;
  Real angley = atan(ypfac) + angleyMCS;

  Real tanglex = tan(anglex);
  Real tangley = tan(angley);

  px = tanglex * (1000.* pfac);
  py = tangley * (1000.* pfac);

  Real directionfac = sqrt(1.0 + xpfac * xpfac + ypfac * ypfac);
  Real zstep = stepsize / directionfac;

  x += zstep * (xpfac + tanglex) / 2.0;
  y += zstep * (ypfac + tangley) / 2.0;
}  

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::IonEnergyLoss
//
// DESCRIPTION
//   Compute ionization energy loss for a particle.  See Bethe-Bloch
//   equation in the Particle Data Book.
//
// PARAMETERS
//   beta:  relativistic beta.
//   Z:     material charge number.
//   A:     atomic number.
//
// RETURNS
//   Real.
//
///////////////////////////////////////////////////////////////////////////

Real Collimator::IonEnergyLoss(Real &beta, Real &Z, Real &A)
{ 
  Real dE, gamma, T_max, m_e=.511, M=938.27, I, arg;
  
  I=Z*10.;
  gamma=1/sqrt(1-beta*beta);                  
  T_max=1e6 * (2*m_e*beta*beta*gamma*gamma)/
    (1 + 2*gamma*m_e/M + Sqr(m_e/M));
  arg=2*(m_e*1e6)*beta*beta*gamma*gamma*T_max/(I*I);
  dE=.307075*Z/A/(beta*beta)*(log(arg)/2. - beta*beta);

  return dE;
  
}  


///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::getKick
//
// DESCRIPTION
//   Generates random, 2D projection of momentum kick.  See Monte 
//   Carlo Methods in PDG, or the K2 code description in the PhD
//   thesis of Nuria Catalan-Lasheras.
//
//
// PARAMETERS
//   t:     magnitude of momentum transfer
//   p:     particle momentum
//   dp_x:  x momentum kick generated
//   dp_x:  y momentum kick generated
//
// RETURNS
//   Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::getKick(Real &t, Real &p, Real &M_nuc, Real &dp_x, 
			 Real &dp_y, Integer &idum)
{ 
  Real va, vb, va2, vb2, r2=10., theta, arg, E;
  Real m=.938;
  
  theta = acos(1 - t/(2*p*p));
  
  while(r2 > 1.)
    {
      va=2.*Collimator::ran1(idum)-1;
      vb=Collimator::ran1(idum)-1;
      va2=va*va;
      vb2=vb*vb;
      r2=va2+vb2;
    }

  dp_x=theta * (2.*va*vb)/r2*1000.;
  dp_y=theta * (va2 - vb2)/r2*1000.;
  
}  

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::getRuth_t
//
// DESCRIPTION
//   Generates random momentum transfer for Rutherford interaction
//   See Catalan-Lasheras
//
// PARAMETERS
//   theta: angle used in coulomb scattering
//   E: particle energy
//   p: particle momentum
//   R: nuclear radius
//   Z: charge number of nucleus
//   idum:  random seed
//   rflag:  set to (0) - just find cross section;  set to (1) - find
//          random momentum transfer
//
// RETURNS
//   Real
//
///////////////////////////////////////////////////////////////////////////

Real Collimator::getRuth_t(Real &theta, Real &E, Real &p, Real &R,
			   Real &Z, Integer &idum, Integer &rflag)
{
  float rfunc(float, float, float);
  float qsimp(float (*rfunc)(float, float, float), float a, float b,
	      float p, float fac1);
  int i, j, k, terms = 15;
  float start = 0.01, finish = 10., f_x = 0., t = 0.;
  float random, A, fac1, A_cut, Const, cnorm, beta, sn, powA; 
  float theta_cut, t_cut, rcross, ifactorial;
  const float gamma = 0.5772157, alpha = 1. / 137.;
  const float hbar_convert = 0.3893858;

  Const = PI * Z * Z * alpha * alpha * hbar_convert *
          E * E / (2. * pow(p, 4.));
  theta_cut = 2.325 * theta;
  sn = sin(theta_cut / 2.);
  t_cut = 4. * p * p * pow(sn, 2);
  fac1 = 1. / 3. * Sqr(R / 0.197);
  A_cut = fac1 * t_cut;
  cnorm = exp(-A_cut) / A_cut + gamma + log(A_cut);

  if(rflag == 0)
  {
    rcross = Const * qsimp(rfunc, theta_cut, PI, p, fac1);
    return rcross / 1000.;
  }

  if(rflag == 1)
  {
    int i_sign = -1;
    ifactorial = 1.;
    powA = 1.;
    for(i = 1; i <= terms; i++)
    {
      ifactorial *= i;
      powA *= A_cut;
      cnorm += i_sign * powA / (i * ifactorial);
      i_sign = - i_sign;
    }
    random = cnorm * Collimator::ran1(idum);
    A = start;
    while(A < finish)
    {
      f_x = exp(-A) / A + gamma + log(A);
      i_sign = -1;
      ifactorial = 1.;
      powA = 1.;
      for(i = 1; i <= terms; i++)
      {
        ifactorial *= i;
        powA *= A;
        f_x += i_sign * powA / (i * ifactorial);
        i_sign = - i_sign;
      }
      f_x -= random;
      if(f_x < 0.001)
      {
        t = A / fac1;
        A = finish + 1.;
      }
      A += 0.01;
    }

    if(t == 0.)
    {
      cout << "WARNING: Could not find momentum transfer (t) for";
      cout <<" Rutherford scattering.  Setting t=0.";
    }

  return t;
  }

  if(rflag == 2)
  {
    beta = p / E;
    sn = sin(theta_cut / 2.);
    rcross = PI * Z * Z * alpha * alpha * hbar_convert /
             (beta * beta * beta * beta * E * E) *
             (1. / (sn * sn) - 1.);
    return rcross / 1000.;
  }

  if(rflag == 3)
  {
    sn = sin(theta_cut / 2.);
    random = (1. / (sn * sn) - 1.) * Collimator::ran1(idum) + 1.;
    t = 2. * p / sqrt(random);
    if(t == 0.)
    {
      cout << "WARNING: Could not find momentum transfer (t) for";
      cout <<" Rutherford scattering.  Setting t=0.";
    }
  return t;
  }
}

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::getRuthJackson
//
// DESCRIPTION
//   Rutherford Scattering from an atomic nucleus.
//   Simple formulation from JD Jackson:
//   Classical Electrodynamics, Chapter 13.
//   Units are cgs, esu.
//
// PARAMETERS
//   stepsize: length (thickness) of scattering material {mm}.
//   Z:        atomic number of scattering material.
//   A:        atomic weight of scattering material.
//   rho:      mass density of scattering material.
//   idum:     random number seed.
//   beta:     v/c.
//
// RETURNS
//   Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::getRuthJackson(Real &stepsize, Real &Z, Real &A, Real &rho,
				Integer &idum, Real &beta, Integer &trackit,
                                Real &rcross, Real &thx, Real &thy)
{
  Real pi = Consts::pi;
  Real hbar = 1.05443e-27;
  Real cvel = Consts::vLight * 100.0;
  Real eesu = 4.80286e-10;
  Real AMU = 1.65979e-24;
  Real aBohr = 5.29172e-09;
  Real AP = 1.007593;
  Real gamma = 1.0 / sqrt(1.0 - beta * beta);

  Real aMax = 1.4 * aBohr / pow(Z, 0.333333);
  Real aMin = 1.4e-13 * pow(A, 0.333333);

  Real N = rho / (A * AMU);
  Real T = stepsize / 10.0;

  Real pmom = gamma * beta * cvel * AP * AMU;

  Real thMin = hbar / (pmom * aMax);
  Real thMax = hbar / (pmom * aMin);

  Real thMin2i = 1.0 / (thMin * thMin);
  Real thMax2i = 1.0 / (thMax * thMax);
  Real th2iDiff = thMin2i - thMax2i;

  Real th2s = 2.0 * log(thMax / thMin) / th2iDiff;
  Real coeff = 2.0 * Z * eesu * eesu / (pmom * beta * cvel);

  Real sigTot = pi * coeff * coeff * th2iDiff;

  Real nColl = N * T * sigTot;

  Real th2Tot = nColl * th2s;

  if(thMin < (2.0 * sqrt(th2Tot))) thMin = 2.0 * sqrt(th2Tot);

  thMin2i = 1.0 / (thMin * thMin);
  th2iDiff = thMin2i - thMax2i;
  rcross = 1.e+24 * pi * coeff * coeff * th2iDiff;
  if(rcross < 0.0) rcross = 0.0;

  thMin *= 1.4;
  thMin2i = 1.0 / (thMin * thMin);
  th2iDiff = thMin2i - thMax2i;

  if(trackit != 0)
  {
    thx = 0.0;
    thy = 0.0;

    if(thMin < thMax)
    {
      Real probrp = Collimator::ran1(idum);
      Real probxy = 2.0 * pi * Collimator::ran1(idum);

      Real denom2 = probrp * th2iDiff + thMax2i;
      Real th = sqrt(1.0 / denom2);

      thx = th * cos(probxy);
      thy = th * sin(probxy);
    }
  }
}  

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//   Collimator::getElast_t
//
// DESCRIPTION
//   Generates random momentum angle for low energy elastic scattering.
//
// PARAMETERS
//   p: particle momentum
//   A: nuclear mass number
//   idum:  random seed.
//
// RETURNS
//   Real.
//
///////////////////////////////////////////////////////////////////////////

Real Collimator::getElast_t(Real &p, Real &A, Integer &idum)
{  
 float bessj0(float x);
 float bessj1(float x); 
 float const c=3e8;
 
 int found=0;
 float R, lamda, E, u, sig, cnorm, theta;
 float random, f_x, x=0., angle_cm=0.;
 float M_nuc, E_cm, p_cm, m=.938, h=4.135e-15/1e9;
 Real t=0.;

 M_nuc = A*.931494;
 E=sqrt(p*p + m*m);
 E_cm = sqrt(m*m + M_nuc*M_nuc + 2*E*M_nuc);             //See PDG.
 p_cm = p*M_nuc/E_cm;
 lamda = h*c/p_cm;
 R=1.4e-15 * pow(A, 1./3.) + lamda;
 u=2*R/lamda*sin(PI/2);
 theta=.0001;

 cnorm= 1./2.*(sqrt(PI)-sqrt(PI)*pow(bessj0(u),2)
	       -sqrt(PI)*pow(bessj1(u),2))/(sqrt(PI));
 
 random = Collimator::ran1(idum);

 while(theta<=PI)
   {
     x = 2.*R*sin(theta/2.)/lamda;
     f_x=1./2./cnorm*(sqrt(PI)-sqrt(PI)*pow(bessj0(x),2)
		      -sqrt(PI)*pow(bessj1(x),2))/(sqrt(PI)) - random;
     if( (f_x > -.01) && (f_x < .01) )
       {
	 angle_cm = theta;
	 theta = PI+1.;
	 found=1;
       }
     theta+=1.768e-3;
   }
 
 if(found==0) cout<<"Warning, never found elastic t.\n";
 t=2.*p_cm*p_cm*(1. - cos(angle_cm));
 return t;
}  

///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//  Collimator::ran1
//
// DESCRIPTION
//    Generates a random real number between 0 and 1.
//
// PARAMETERS
//    idum:  Seed number.  Should be a negative integer.
//
// RETURNS
//    Random number.
//
///////////////////////////////////////////////////////////////////////////

Real Collimator::ran1(Integer &idum)
{
  int j;
  long k;
  static long iy=0;
  static long iv[NTAB];
  Real temp;

  if(idum <= 0 || !iy)
  {
    if (-(idum) < 1) idum=1;
    else idum = -(idum);
    for (j=NTAB+7; j>=0; j--)
    {
      k=(idum)/IQ;
      idum=IA*(idum-k*IQ) - IR*k;
      if (idum < 0) idum += IM;
      if (j< NTAB) iv[j] = idum;
    }
    iy=iv[0];
  }
 
  k=(idum)/IQ;
  idum=IA*(idum-k*IQ)-IR*k;
  if (idum < 0) idum += IM;
  j=iy/NDIV;
  iy=iv[j];
  iv[j]=idum;
  if ((temp=AM*iy) > RNMX) return RNMX; 
  else return temp; 

// The following is a standard way of getting the random number in SuperCode
// Changes made by A. Shishlo
// return MathLib::ran1(Injection::randSeed);
}  


///////////////////////////////////////////////////////////////////////////
// 
// NAME 
// Collimate::_getMaterialParams(Integer &_ma)
// 
// DESCRIPTION 
// Initializes table of elestic and inelastic total cross sections vs.
// energy for the user-specified material.  Tables generated with 
// MCNPX for elastic and inelastic cross sections.  Rutherfod
// cross section are calculated via a formula in Particles and
// Nuclei, by Pohv.
//
// All Rutherford cross sections below 20 MeV are set to zero.  
// MCS/Rutherford is not well-behaved below this value.  Particles 
// below 20 MeV are considered lost in collimation routine.
//
// PARAMETERS 
// _ma   - the material number  
// 
// RETURNS 
// Nothing.  
//
///////////////////////////////////////////////////////////////////////////

Void Collimate::_getMaterialParams(Integer &_ma)
{ 

  crossEn(1)=0.0005; crossEn(2)=.001; crossEn(3)=.0015;
  crossEn(4)=0.002; crossEn(5)=.0025; crossEn(6)=.003;
  crossEn(7)=0.0035; crossEn(8)=.004; crossEn(9)=.0045;
  crossEn(10)=0.005; crossEn(11)=.0055; crossEn(12)=.006;
  crossEn(13)=0.0065; crossEn(14)=.007; crossEn(15)=.0075;
  crossEn(16)=0.008; crossEn(17)=.009; crossEn(18)=.01;
  crossEn(19)=0.011; crossEn(20)=.012; crossEn(21)=.013;
  crossEn(22)=0.014; crossEn(23)=.015; crossEn(24)=.0175;
  crossEn(25)=0.02; crossEn(26)=.0225; crossEn(27)=.025;
  crossEn(28)=.030; crossEn(29)=0.035; crossEn(30)=.040;
  crossEn(31)=.045; crossEn(32)=.050; crossEn(33)=.0510;
  crossEn(34)=.055; crossEn(35)=.060; crossEn(36)=.070;
  crossEn(37)=.080; crossEn(38)=.090; crossEn(39)=.100;
  crossEn(40)=.110; crossEn(41)=.120; crossEn(42)=.140;
  crossEn(43)=.160; crossEn(44)=.180; crossEn(45)=.200;
  crossEn(46)=.225; crossEn(47)=.250; crossEn(48)=.275;
  crossEn(49)=.300; crossEn(50)=.325; crossEn(51)=.350;
  crossEn(52)=.375; crossEn(53)=.400; crossEn(54)=.500;
  crossEn(55)=.700; crossEn(56)=1.000; crossEn(57)=1.500;
  
  if(_ma==1)
    {
      _Z=6.; _A=12.011; _density=2.265; _radlength=18.8;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=.321;
      crossEl(7)=.335; crossEl(8)=.349; crossEl(9)=.363;
      crossEl(10)=.377; crossEl(11)=.386; crossEl(12)=.395;
      crossEl(13)=.404; crossEl(14)=.413; crossEl(15)=.423;
      crossEl(16)=.434; crossEl(17)=.456; crossEl(18)=.479;
      crossEl(19)=.509; crossEl(20)=.539; crossEl(21)=.569;
      crossEl(22)=.598; crossEl(23)=.628; crossEl(24)=.685;
      crossEl(25)=.743; crossEl(26)=.783; crossEl(27)=.822;
      crossEl(28)=.878; crossEl(29)=.915; crossEl(30)=.938;
      crossEl(31)=.813; crossEl(32)=.688; crossEl(33)=.678;
      crossEl(34)=.641; crossEl(35)=.594; crossEl(36)=.515;
      crossEl(37)=.448; crossEl(38)=.392; crossEl(39)=.345;
      crossEl(40)=.305; crossEl(41)=.272; crossEl(42)=.214;
      crossEl(43)=.167; crossEl(44)=.138; crossEl(45)=.117;
      crossEl(46)=.098; crossEl(47)=.085; crossEl(48)=.077;
      crossEl(49)=.072; crossEl(50)=.068; crossEl(51)=.067;
      crossEl(52)=.066; crossEl(53)=.067; crossEl(54)=.077;
      crossEl(55)=.09; crossEl(56)=.102; crossEl(57)=.108;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=0.; crossInel(9)=0.;
      crossInel(10)=.013; crossInel(11)=.078; crossInel(12)=.131;
      crossInel(13)=.175; crossInel(14)=.212; crossInel(15)=.244;
      crossInel(16)=.271; crossInel(17)=.314; crossInel(18)=.346;
      crossInel(19)=.370; crossInel(20)=.389; crossInel(21)=.403;
      crossInel(22)=.413; crossInel(23)=.421; crossInel(24)=.432;
      crossInel(25)=.434; crossInel(26)=.432; crossInel(27)=.427;
      crossInel(28)=.412; crossInel(29)=.394; crossInel(30)=.376;
      crossInel(31)=.359; crossInel(32)=.344; crossInel(33)=.288;
      crossInel(34)=.281; crossInel(35)=.272; crossInel(36)=.257;
      crossInel(37)=.245; crossInel(38)=.235; crossInel(39)=.228;
      crossInel(40)=.223; crossInel(41)=.220; crossInel(42)=.223;
      crossInel(43)=.213; crossInel(44)=.212; crossInel(45)=.212;
      crossInel(46)=.212; crossInel(47)=.213; crossInel(48)=.214;
      crossInel(49)=.216; crossInel(50)=.217; crossInel(51)=.219;
      crossInel(52)=.221; crossInel(53)=.223; crossInel(54)=.229;
      crossInel(55)=.246; crossInel(56)=.261; crossInel(57)=.261;
      
      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.00167; crossRuth(26)=0.00190; crossRuth(27)=0.00211;
      crossRuth(28)=0.00250; crossRuth(29)=0.00285; crossRuth(30)=0.00316;
      crossRuth(31)=0.00344; crossRuth(32)=0.00369; crossRuth(33)=0.00373;
      crossRuth(34)=0.00391; crossRuth(35)=0.00411; crossRuth(36)=0.00446;
      crossRuth(37)=0.00476; crossRuth(38)=0.00500; crossRuth(39)=0.00522;
      crossRuth(40)=0.00540; crossRuth(41)=0.00557; crossRuth(42)=0.00584;
      crossRuth(43)=0.00606; crossRuth(44)=0.00624; crossRuth(45)=0.00640;
      crossRuth(46)=0.00656; crossRuth(47)=0.00670; crossRuth(48)=0.00680;
      crossRuth(49)=0.00690; crossRuth(50)=0.00698; crossRuth(51)=0.00706;
      crossRuth(52)=0.00712; crossRuth(53)=0.00718; crossRuth(54)=0.00735;
      crossRuth(55)=0.00757; crossRuth(56)=0.00773; crossRuth(57)=0.00785;
    }
  
  if(_ma==2)
    {
      _Z=13.; _A=26.982; _density=2.7; _radlength=8.9;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=.091;
      crossEl(7)=.164; crossEl(8)=.237; crossEl(9)=.311;
      crossEl(10)=.384; crossEl(11)=.428; crossEl(12)=.473;
      crossEl(13)=.518; crossEl(14)=.562; crossEl(15)=.607;
      crossEl(16)=.623; crossEl(17)=.654; crossEl(18)=.686;
      crossEl(19)=.692; crossEl(20)=.699; crossEl(21)=.701;
      crossEl(22)=.700; crossEl(23)=.700; crossEl(24)=.707;
      crossEl(25)=.714; crossEl(26)=.737; crossEl(27)=.761;
      crossEl(28)=.832; crossEl(29)=.905; crossEl(30)=.976;
      crossEl(31)=.993; crossEl(32)=1.010; crossEl(33)=1.007;
      crossEl(34)=.994; crossEl(35)=.979; crossEl(36)=.921;
      crossEl(37)=.854; crossEl(38)=.786; crossEl(39)=.720;
      crossEl(40)=.658; crossEl(41)=.601; crossEl(42)=.490;
      crossEl(43)=.382; crossEl(44)=.320; crossEl(45)=.274;
      crossEl(46)=.231; crossEl(47)=.202; crossEl(48)=.181;
      crossEl(49)=.167; crossEl(50)=.159; crossEl(51)=.153;
      crossEl(52)=.151; crossEl(53)=.151; crossEl(54)=.170;
      crossEl(55)=.198; crossEl(56)=.220; crossEl(57)=.232;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=.123; crossInel(6)=.280;
      crossInel(7)=.391; crossInel(8)=.474; crossInel(9)=.537;
      crossInel(10)=.587; crossInel(11)=.626; crossInel(12)=.658;
      crossInel(13)=.684; crossInel(14)=.705; crossInel(15)=.722;
      crossInel(16)=.736; crossInel(17)=.757; crossInel(18)=.771;
      crossInel(19)=.780; crossInel(20)=.785; crossInel(21)=.787;
      crossInel(22)=.787; crossInel(23)=.786; crossInel(24)=.777;
      crossInel(25)=.764; crossInel(26)=.749; crossInel(27)=.732;
      crossInel(28)=.699; crossInel(29)=.667; crossInel(30)=.638;
      crossInel(31)=.611; crossInel(32)=.587; crossInel(33)=.530;
      crossInel(34)=.518; crossInel(35)=.503; crossInel(36)=.478;
      crossInel(37)=.456; crossInel(38)=.437; crossInel(39)=.422;
      crossInel(40)=.410; crossInel(41)=.402; crossInel(42)=.397;
      crossInel(43)=.394; crossInel(44)=.393; crossInel(45)=.392;
      crossInel(46)=.393; crossInel(47)=.395; crossInel(48)=.397;
      crossInel(49)=.399; crossInel(50)=.402; crossInel(51)=.405;
      crossInel(52)=.408; crossInel(53)=.411; crossInel(54)=.415;
      crossInel(55)=.438; crossInel(56)=.457; crossInel(57)=.460;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.00311; crossRuth(26)=0.00379; crossRuth(27)=0.00448;
      crossRuth(28)=0.00586; crossRuth(29)=0.00718; crossRuth(30)=0.00843;
      crossRuth(31)=0.00960; crossRuth(32)=0.0107; crossRuth(33)=0.0109;
      crossRuth(34)=0.0117; crossRuth(35)=0.0126; crossRuth(36)=0.0143;
      crossRuth(37)=0.0157; crossRuth(38)=0.0170; crossRuth(39)=0.0181;
      crossRuth(40)=0.0191; crossRuth(41)=0.0200; crossRuth(42)=0.0215;
      crossRuth(43)=0.0227; crossRuth(44)=0.0238; crossRuth(45)=0.0246;
      crossRuth(46)=0.0256; crossRuth(47)=0.0264; crossRuth(48)=0.0270;
      crossRuth(49)=0.0276; crossRuth(50)=0.0281; crossRuth(51)=0.0286;
      crossRuth(52)=0.0290; crossRuth(53)=0.0293; crossRuth(54)=0.0304;
      crossRuth(55)=0.0317; crossRuth(56)=0.0327; crossRuth(57)=0.0335;
    }
  
  if(_ma==3)
    {
      _Z=26.; _A=55.845; _density=7.87; _radlength=1.76;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=.001;
      crossEl(7)=.015; crossEl(8)=.030; crossEl(9)=.045;
      crossEl(10)=.059; crossEl(11)=.109; crossEl(12)=.159;
      crossEl(13)=.209; crossEl(14)=.259; crossEl(15)=.309;
      crossEl(16)=.364; crossEl(17)=.474; crossEl(18)=.584;
      crossEl(19)=.668; crossEl(20)=.753; crossEl(21)=.819;
      crossEl(22)=.867; crossEl(23)=.915; crossEl(24)=.964;
      crossEl(25)=1.013; crossEl(26)=1.030; crossEl(27)=1.047;
      crossEl(28)=1.063; crossEl(29)=1.075; crossEl(30)=1.098;
      crossEl(31)=1.121; crossEl(32)=1.143; crossEl(33)=1.153;
      crossEl(34)=1.192; crossEl(35)=1.240; crossEl(36)=1.293;
      crossEl(37)=1.306; crossEl(38)=1.289; crossEl(39)=1.250;
      crossEl(40)=1.198; crossEl(41)=1.137; crossEl(42)=.984;
      crossEl(43)=.777; crossEl(44)=.669; crossEl(45)=.582;
      crossEl(46)=.499; crossEl(47)=.438; crossEl(48)=.393;
      crossEl(49)=.361; crossEl(50)=.340; crossEl(51)=.325;
      crossEl(52)=.317; crossEl(53)=.314; crossEl(54)=.346;
      crossEl(55)=.389; crossEl(56)=.425; crossEl(57)=.444;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=.017; crossInel(9)=.174;
      crossInel(10)=.301; crossInel(11)=.404; crossInel(12)=.490;
      crossInel(13)=.562; crossInel(14)=.623; crossInel(15)=.676;
      crossInel(16)=.721; crossInel(17)=.795; crossInel(18)=.852;
      crossInel(19)=.896; crossInel(20)=.931; crossInel(21)=.959;
      crossInel(22)=.981; crossInel(23)=.999; crossInel(24)=1.028;
      crossInel(25)=1.044; crossInel(26)=1.05; crossInel(27)=1.05;
      crossInel(28)=1.04; crossInel(29)=1.021; crossInel(30)=.999;
      crossInel(31)=.976; crossInel(32)=.952; crossInel(33)=.920;
      crossInel(34)=.899; crossInel(35)=.873; crossInel(36)=.834;
      crossInel(37)=.790; crossInel(38)=.753; crossInel(39)=.719;
      crossInel(40)=.698; crossInel(41)=.685; crossInel(42)=.672;
      crossInel(43)=.677; crossInel(44)=.674; crossInel(45)=.673;
      crossInel(46)=.674; crossInel(47)=.675; crossInel(48)=.678;
      crossInel(49)=.681; crossInel(50)=.685; crossInel(51)=.689;
      crossInel(52)=.693; crossInel(53)=.697; crossInel(54)=.679;
      crossInel(55)=.710; crossInel(56)=.735; crossInel(57)=.744;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.00349; crossRuth(26)=0.00472; crossRuth(27)=0.00609;
      crossRuth(28)=0.00913; crossRuth(29)=0.0124; crossRuth(30)=0.0158;
      crossRuth(31)=0.0191; crossRuth(32)=0.0224; crossRuth(33)=0.0230;
      crossRuth(34)=0.0255; crossRuth(35)=0.0286; crossRuth(36)=0.0343;
      crossRuth(37)=0.0395; crossRuth(38)=0.0443; crossRuth(39)=0.0485;
      crossRuth(40)=0.0524; crossRuth(41)=0.0560; crossRuth(42)=0.0622;
      crossRuth(43)=0.0675; crossRuth(44)=0.0720; crossRuth(45)=0.0759;
      crossRuth(46)=0.0801; crossRuth(47)=0.0836; crossRuth(48)=0.0867;
      crossRuth(49)=0.0894; crossRuth(50)=0.0918; crossRuth(51)=0.0939;
      crossRuth(52)=0.0958; crossRuth(53)=0.0975; crossRuth(54)=0.103;
      crossRuth(55)=0.109; crossRuth(56)=0.114; crossRuth(57)=0.118;
    }
  
  if(_ma==4)
    {
      _Z=29.; _A=63.546; _density=8.96; _radlength=1.43;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=.0;
      crossEl(7)=.008; crossEl(8)=.017; crossEl(9)=.025;
      crossEl(10)=.033; crossEl(11)=.077; crossEl(12)=.121;
      crossEl(13)=.165; crossEl(14)=.208; crossEl(15)=.252;
      crossEl(16)=.310; crossEl(17)=.426; crossEl(18)=.541;
      crossEl(19)=.634; crossEl(20)=.726; crossEl(21)=.800;
      crossEl(22)=.855; crossEl(23)=.910; crossEl(24)=.971;
      crossEl(25)=1.033; crossEl(26)=1.058; crossEl(27)=1.083;
      crossEl(28)=1.113; crossEl(29)=1.132; crossEl(30)=1.156;
      crossEl(31)=1.161; crossEl(32)=1.165; crossEl(33)=1.175;
      crossEl(34)=1.216; crossEl(35)=1.268; crossEl(36)=1.337;
      crossEl(37)=1.368; crossEl(38)=1.368; crossEl(39)=1.343;
      crossEl(40)=1.300; crossEl(41)=1.245; crossEl(42)=1.092;
      crossEl(43)=.869; crossEl(44)=.752; crossEl(45)=.657;
      crossEl(46)=.565; crossEl(47)=.496; crossEl(48)=.446;
      crossEl(49)=.410; crossEl(50)=.384; crossEl(51)=.368;
      crossEl(52)=.358; crossEl(53)=.353; crossEl(54)=.385;
      crossEl(55)=.429; crossEl(56)=.468; crossEl(57)=.488;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=.063; crossInel(9)=.225;
      crossInel(10)=.354; crossInel(11)=.460; crossInel(12)=.548;
      crossInel(13)=.622; crossInel(14)=.685; crossInel(15)=.739;
      crossInel(16)=.786; crossInel(17)=.862; crossInel(18)=.920;
      crossInel(19)=.966; crossInel(20)=1.003; crossInel(21)=1.032;
      crossInel(22)=1.055; crossInel(23)=1.073; crossInel(24)=1.104;
      crossInel(25)=1.121; crossInel(26)=1.128; crossInel(27)=1.128;
      crossInel(28)=1.118; crossInel(29)=1.100; crossInel(30)=1.077;
      crossInel(31)=1.053; crossInel(32)=1.029; crossInel(33)=.986;
      crossInel(34)=.966; crossInel(35)=.942; crossInel(36)=.902;
      crossInel(37)=.859; crossInel(38)=.822; crossInel(39)=.788;
      crossInel(40)=.765; crossInel(41)=.749; crossInel(42)=.732;
      crossInel(43)=.736; crossInel(44)=.733; crossInel(45)=.732;
      crossInel(46)=.732; crossInel(47)=.734; crossInel(48)=.737;
      crossInel(49)=.740; crossInel(50)=.744; crossInel(51)=.748;
      crossInel(52)=.752; crossInel(53)=.756; crossInel(54)=.736;
      crossInel(55)=.768; crossInel(56)=.794; crossInel(57)=.805;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.00326; crossRuth(26)=0.00452; crossRuth(27)=0.00595;
      crossRuth(28)=0.00920; crossRuth(29)=0.0128; crossRuth(30)=0.0165;
      crossRuth(31)=0.0203; crossRuth(32)=0.0241; crossRuth(33)=0.0248;
      crossRuth(34)=0.0278; crossRuth(35)=0.0314; crossRuth(36)=0.0381;
      crossRuth(37)=0.0444; crossRuth(38)=0.0501; crossRuth(39)=0.0553;
      crossRuth(40)=0.0601; crossRuth(41)=0.0644; crossRuth(42)=0.0721;
      crossRuth(43)=0.0786; crossRuth(44)=0.0843; crossRuth(45)=0.0891;
      crossRuth(46)=0.0944; crossRuth(47)=0.0989; crossRuth(48)=0.103;
      crossRuth(49)=0.106; crossRuth(50)=0.109; crossRuth(51)=0.112;
      crossRuth(52)=0.114; crossRuth(53)=0.116; crossRuth(54)=0.123;
      crossRuth(55)=0.131; crossRuth(56)=0.138; crossRuth(57)=0.143;
    }
  

 if(_ma==5)
    {
      _Z=73.; _A=180.9479; _density=16.6; _radlength=.411;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0; crossEl(2)=0; crossEl(3)=0; 
      crossEl(4)=0; crossEl(5)=0; crossEl(6)=0;
      crossEl(7)=0; crossEl(8)=0; crossEl(9)=0;
      crossEl(10)=0; crossEl(11)=0; crossEl(12)=0;
      crossEl(13)=0; crossEl(14)=0; crossEl(15)=0;
      crossEl(16)=.004; crossEl(17)=.011; crossEl(18)=.019;
      crossEl(19)=.072; crossEl(20)=.126; crossEl(21)=.207;
      crossEl(22)=.316; crossEl(23)=.425; crossEl(24)=.704;
      crossEl(25)=.982; crossEl(26)=1.124; crossEl(27)=1.266;
      crossEl(28)=1.401; crossEl(29)=1.506; crossEl(30)=1.586;
      crossEl(31)=1.593; crossEl(32)=1.600; crossEl(33)=1.608;
      crossEl(34)=1.639; crossEl(35)=1.679; crossEl(36)=1.773;
      crossEl(37)=1.884; crossEl(38)=2.001; crossEl(39)=2.109;
      crossEl(40)=2.198; crossEl(41)=2.261; crossEl(42)=2.255;
      crossEl(43)=1.984; crossEl(44)=1.831; crossEl(45)=1.679;
      crossEl(46)=1.506; crossEl(47)=1.359; crossEl(48)=1.240;
      crossEl(49)=1.146; crossEl(50)=1.074; crossEl(51)=1.022;
      crossEl(52)=.985; crossEl(53)=.963; crossEl(54)=.997;
      crossEl(55)=1.053; crossEl(56)=1.119; crossEl(57)=1.186;
      
      crossInel(1)=0; crossInel(2)=0; crossInel(3)=0;
      crossInel(4)=0; crossInel(5)=0; crossInel(6)=0;
      crossInel(7)=0; crossInel(8)=0; crossInel(9)=0;
      crossInel(10)=0; crossInel(11)=0; crossInel(12)=0;
      crossInel(13)=0; crossInel(14)=0; crossInel(15)=0;
      crossInel(16)=.012; crossInel(17)=.286; crossInel(18)=.505;
      crossInel(19)=.685; crossInel(20)=.833; crossInel(21)=.958;
      crossInel(22)=1.065; crossInel(23)=1.156; crossInel(24)=1.334;
      crossInel(25)=1.462; crossInel(26)=1.557; crossInel(27)=1.628;
      crossInel(28)=1.723; crossInel(29)=1.779; crossInel(30)=1.810;
      crossInel(31)=1.825; crossInel(32)=1.831; crossInel(33)=1.824;
      crossInel(34)=1.823; crossInel(35)=1.821; crossInel(36)=1.807;
      crossInel(37)=1.784; crossInel(38)=1.761; crossInel(39)=1.736;
      crossInel(40)=1.705; crossInel(41)=1.676; crossInel(42)=1.629;
      crossInel(43)=1.604; crossInel(44)=1.591; crossInel(45)=1.578;
      crossInel(46)=1.574; crossInel(47)=1.568; crossInel(48)=1.564;
      crossInel(49)=1.560; crossInel(50)=1.555; crossInel(51)=1.551;
      crossInel(52)=1.546; crossInel(53)=1.541; crossInel(54)=1.549;
      crossInel(55)=1.599; crossInel(56)=1.641; crossInel(57)=1.666;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.000761; crossRuth(26)=0.00140; crossRuth(27)=0.00234;
      crossRuth(28)=0.00522; crossRuth(29)=0.00953; crossRuth(30)=0.0152;
      crossRuth(31)=0.0221; crossRuth(32)=0.0300; crossRuth(33)=0.0316;
      crossRuth(34)=0.0387; crossRuth(35)=0.0480; crossRuth(36)=0.0680;
      crossRuth(37)=0.0887; crossRuth(38)=0.110; crossRuth(39)= 0.131;
      crossRuth(40)=0.151; crossRuth(41)=0.170; crossRuth(42)=0.207;
      crossRuth(43)=0.241; crossRuth(44)=0.271; crossRuth(45)=0.299;
      crossRuth(46)=0.330; crossRuth(47)=0.357; crossRuth(48)=0.382;
      crossRuth(49)=0.404; crossRuth(50)=0.423; crossRuth(51)=0.441;
      crossRuth(52)=0.457; crossRuth(53)=0.472; crossRuth(54)=0.520;
      crossRuth(55)=0.581; crossRuth(56)=0.631; crossRuth(57)=0.671;
    }
  



  if(_ma==6)
    {
      _Z=74.; _A=183.84; _density=19.3; _radlength=.35;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=0.;
      crossEl(7)=0.; crossEl(8)=0.; crossEl(9)=0.;
      crossEl(10)=0.; crossEl(11)=0.; crossEl(12)=0.;
      crossEl(13)=0.; crossEl(14)=0.; crossEl(15)=0.;
      crossEl(16)=.003; crossEl(17)=.010; crossEl(18)=.017;
      crossEl(19)=.067; crossEl(20)=.118; crossEl(21)=.197;
      crossEl(22)=.305; crossEl(23)=.412; crossEl(24)=.694;
      crossEl(25)=.975; crossEl(26)=1.121; crossEl(27)=1.267;
      crossEl(28)=1.404; crossEl(29)=1.511; crossEl(30)=1.591;
      crossEl(31)=1.600; crossEl(32)=1.610; crossEl(33)=1.618;
      crossEl(34)=1.649; crossEl(35)=1.689; crossEl(36)=1.782;
      crossEl(37)=1.893; crossEl(38)=2.010; crossEl(39)=2.120;
      crossEl(40)=2.211; crossEl(41)=2.276; crossEl(42)=2.275;
      crossEl(43)=2.006; crossEl(44)=1.852; crossEl(45)=1.700;
      crossEl(46)=1.526; crossEl(47)=1.378; crossEl(48)=1.258;
      crossEl(49)=1.163; crossEl(50)=1.090; crossEl(51)=1.037;
      crossEl(52)=1.000; crossEl(53)=.977; crossEl(54)=1.011;
      crossEl(55)=1.067; crossEl(56)=1.134; crossEl(57)=1.171;
	
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=0.; crossInel(9)=0.;
      crossInel(10)=0.; crossInel(11)=0.; crossInel(12)=0.;
      crossInel(13)=0.; crossInel(14)=0.; crossInel(15)=0.;
      crossInel(16)=0.; crossInel(17)=.265; crossInel(18)=.489;
      crossInel(19)=.672; crossInel(20)=.823; crossInel(21)=.951;
      crossInel(22)=1.059; crossInel(23)=1.152; crossInel(24)=1.334;
      crossInel(25)=1.465; crossInel(26)=1.562; crossInel(27)=1.635;
      crossInel(28)=1.733; crossInel(29)=1.790; crossInel(30)=1.823;
      crossInel(31)=1.840; crossInel(32)=1.846; crossInel(33)=1.840;
      crossInel(34)=1.839; crossInel(35)=1.838; crossInel(36)=1.825;
      crossInel(37)=1.804; crossInel(38)=1.782; crossInel(39)=1.759;
      crossInel(40)=1.728; crossInel(41)=1.700; crossInel(42)=1.653;
      crossInel(43)=1.626; crossInel(44)=1.612; crossInel(45)=1.599;
      crossInel(46)=1.594; crossInel(47)=1.588; crossInel(48)=1.583;
      crossInel(49)=1.578; crossInel(50)=1.573; crossInel(51)=1.568;
      crossInel(52)=1.562; crossInel(53)=1.557; crossInel(54)=1.568;
      crossInel(55)=1.618; crossInel(56)=1.659; crossInel(57)=1.686;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.000731; crossRuth(26)=0.00136; crossRuth(27)=0.00227;
      crossRuth(28)=0.00511; crossRuth(29)=0.00938; crossRuth(30)=0.0150;
      crossRuth(31)=0.0219; crossRuth(32)=0.0298; crossRuth(33)=0.0315;
      crossRuth(34)=0.0385; crossRuth(35)=0.0479; crossRuth(36)=0.0680;
      crossRuth(37)=0.0891; crossRuth(38)=0.110; crossRuth(39)=0.131;
      crossRuth(40)=0.152; crossRuth(41)=0.172; crossRuth(42)=0.210;
      crossRuth(43)=0.244; crossRuth(44)=0.275; crossRuth(45)=0.303;
      crossRuth(46)=0.335; crossRuth(47)=0.363; crossRuth(48)=0.388;
      crossRuth(49)=0.411; crossRuth(50)=0.431; crossRuth(51)=0.449;
      crossRuth(52)=0.466; crossRuth(53)=0.481; crossRuth(54)=0.530;
      crossRuth(55)=0.593; crossRuth(56)=0.644; crossRuth(57)=0.686;
    }

  if(_ma==7)
    {
      _Z=78.; _A=195.08; _density=21.45; _radlength=.305;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=0.;
      crossEl(7)=0.; crossEl(8)=0.; crossEl(9)=0.;
      crossEl(10)=0.; crossEl(11)=0.; crossEl(12)=0.;
      crossEl(13)=0.; crossEl(14)=0.; crossEl(15)=0.;
      crossEl(16)=.002; crossEl(17)=.006; crossEl(18)=.010;
      crossEl(19)=.051; crossEl(20)=.091; crossEl(21)=.161;
      crossEl(22)=.261; crossEl(23)=.361; crossEl(24)=.654;
      crossEl(25)=.947; crossEl(26)=1.108; crossEl(27)=1.269;
      crossEl(28)=1.416; crossEl(29)=1.530; crossEl(30)=1.614;
      crossEl(31)=1.632; crossEl(32)=1.651; crossEl(33)=1.659;
      crossEl(34)=1.691; crossEl(35)=1.731; crossEl(36)=1.821;
      crossEl(37)=1.931; crossEl(38)=2.051; crossEl(39)=2.166;
      crossEl(40)=2.265; crossEl(41)=2.340; crossEl(42)=2.365;
      crossEl(43)=2.096; crossEl(44)=1.944; crossEl(45)=1.790;
      crossEl(46)=1.613; crossEl(47)=1.460; crossEl(48)=1.335;
      crossEl(49)=1.235; crossEl(50)=1.159; crossEl(51)=1.103;
      crossEl(52)=1.063; crossEl(53)=1.038; crossEl(54)=1.072;
      crossEl(55)=1.129; crossEl(56)=1.198; crossEl(57)=1.236;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=0.; crossInel(9)=0.;
      crossInel(10)=0.; crossInel(11)=0.; crossInel(12)=0.;
      crossInel(13)=0.; crossInel(14)=0.; crossInel(15)=0.;
      crossInel(16)=0.; crossInel(17)=.149; crossInel(18)=.392;
      crossInel(19)=.591; crossInel(20)=.757; crossInel(21)=.896;
      crossInel(22)=1.015; crossInel(23)=1.117; crossInel(24)=1.317;
      crossInel(25)=1.462; crossInel(26)=1.570; crossInel(27)=1.652;
      crossInel(28)=1.763; crossInel(29)=1.830; crossInel(30)=1.870;
      crossInel(31)=1.893; crossInel(32)=1.904; crossInel(33)=1.967;
      crossInel(34)=1.953; crossInel(35)=1.935; crossInel(36)=1.909;
      crossInel(37)=1.878; crossInel(38)=1.842; crossInel(39)=1.806;
      crossInel(40)=1.768; crossInel(41)=1.734; crossInel(42)=1.682;
      crossInel(43)=1.664; crossInel(44)=1.656; crossInel(45)=1.649;
      crossInel(46)=1.647; crossInel(47)=1.647; crossInel(48)=1.647;
      crossInel(49)=1.648; crossInel(50)=1.649; crossInel(51)=1.650;
      crossInel(52)=1.651; crossInel(53)=1.653; crossInel(54)=1.644;
      crossInel(55)=1.695; crossInel(56)=1.738; crossInel(57)=1.766;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.000627; crossRuth(26)=0.00119; crossRuth(27)=0.00203;
      crossRuth(28)=0.00471; crossRuth(29)=0.00883; crossRuth(30)=0.0144;
      crossRuth(31)=0.0212; crossRuth(32)=0.0292; crossRuth(33)=0.0309;
      crossRuth(34)=0.0381; crossRuth(35)=0.0478; crossRuth(36)=0.0686;
      crossRuth(37)=0.0907; crossRuth(38)=0.113; crossRuth(39)=0.136;
      crossRuth(40)=0.158; crossRuth(41)=0.179; crossRuth(42)=0.219;
      crossRuth(43)=0.256; crossRuth(44)=0.290; crossRuth(45)=0.321;
      crossRuth(46)=0.356; crossRuth(47)=0.387; crossRuth(48)=0.414;
      crossRuth(49)=0.439; crossRuth(50)=0.461; crossRuth(51)=0.482;
      crossRuth(52)=0.500; crossRuth(53)=0.517; crossRuth(54)=0.571;
      crossRuth(55)=0.641; crossRuth(56)=0.698; crossRuth(57)=0.745;
    }

  if(_ma==8)
    {
      _Z=82.; _A=207.2; _density=11.35; _radlength=.56;
      _R=0.94*Pow(_A, 1./3.);

      crossEl(1)=0.; crossEl(2)=0.; crossEl(3)=0.; 
      crossEl(4)=0.; crossEl(5)=0.; crossEl(6)=0.;
      crossEl(7)=0.; crossEl(8)=0.; crossEl(9)=0.;
      crossEl(10)=0.; crossEl(11)=0.; crossEl(12)=0.;
      crossEl(13)=0.; crossEl(14)=0.; crossEl(15)=0.;
      crossEl(16)=.001; crossEl(17)=.004; crossEl(18)=.007;
      crossEl(19)=.040; crossEl(20)=.074; crossEl(21)=.138;
      crossEl(22)=.231; crossEl(23)=.325; crossEl(24)=.625;
      crossEl(25)=.925; crossEl(26)=1.097; crossEl(27)=1.270;
      crossEl(28)=1.426; crossEl(29)=1.545; crossEl(30)=1.633;
      crossEl(31)=1.659; crossEl(32)=1.684; crossEl(33)=1.692;
      crossEl(34)=1.725; crossEl(35)=1.765; crossEl(36)=1.853;
      crossEl(37)=1.962; crossEl(38)=2.083; crossEl(39)=2.203;
      crossEl(40)=2.309; crossEl(41)=2.392; crossEl(42)=2.423;
      crossEl(43)=2.170; crossEl(44)=2.020; crossEl(45)=1.865;
      crossEl(46)=1.685; crossEl(47)=1.528; crossEl(48)=1.399;
      crossEl(49)=1.296; crossEl(50)=1.216; crossEl(51)=1.158;
      crossEl(52)=1.116; crossEl(53)=1.090; crossEl(54)=1.124;
      crossEl(55)=1.181; crossEl(56)=1.251; crossEl(57)=1.291;
      
      crossInel(1)=0.; crossInel(2)=0.; crossInel(3)=0.;
      crossInel(4)=0.; crossInel(5)=0.; crossInel(6)=0.;
      crossInel(7)=0.; crossInel(8)=0.; crossInel(9)=0.;
      crossInel(10)=0.; crossInel(11)=0.; crossInel(12)=0.;
      crossInel(13)=0.; crossInel(14)=0.; crossInel(15)=0.;
      crossInel(16)=0.; crossInel(17)=.090; crossInel(18)=.346;
      crossInel(19)=.556; crossInel(20)=.731; crossInel(21)=.878;
      crossInel(22)=1.004; crossInel(23)=1.112; crossInel(24)=1.325;
      crossInel(25)=1.479; crossInel(26)=1.594; crossInel(27)=1.682;
      crossInel(28)=1.803; crossInel(29)=1.877; crossInel(30)=1.922;
      crossInel(31)=1.948; crossInel(32)=1.962; crossInel(33)=2.074;
      crossInel(34)=2.048; crossInel(35)=2.015; crossInel(36)=1.979;
      crossInel(37)=1.939; crossInel(38)=1.892; crossInel(39)=1.845;
      crossInel(40)=1.801; crossInel(41)=1.762; crossInel(42)=1.706;
      crossInel(43)=1.695; crossInel(44)=1.691; crossInel(45)=1.690;
      crossInel(46)=1.691; crossInel(47)=1.695; crossInel(48)=1.699;
      crossInel(49)=1.705; crossInel(50)=1.712; crossInel(51)=1.718;
      crossInel(52)=1.725; crossInel(53)=1.733; crossInel(54)=1.711;
      crossInel(55)=1.763; crossInel(56)=1.807; crossInel(57)=1.836;

      crossRuth(1)=0.0; crossRuth(2)=0.0; crossRuth(3)=0.0;
      crossRuth(4)=0.0; crossRuth(5)=0.0; crossRuth(6)=0.0;
      crossRuth(7)=0.0; crossRuth(8)=0.0; crossRuth(9)=0.0;
      crossRuth(10)=0.0; crossRuth(11)=0.0; crossRuth(12)=0.0;
      crossRuth(13)=0.0; crossRuth(14)=0.0; crossRuth(15)=0.0;
      crossRuth(16)=0.0; crossRuth(17)=0.0; crossRuth(18)=0.0;
      crossRuth(19)=0.0; crossRuth(20)=0.0; crossRuth(21)=0.0;
      crossRuth(22)=0.0; crossRuth(23)=0.0; crossRuth(24)=0.0;
      crossRuth(25)=0.000528; crossRuth(26)=0.00103; crossRuth(27)=0.00179;
      crossRuth(28)=0.00427; crossRuth(29)=0.00820; crossRuth(30)=0.0136;
      crossRuth(31)=0.0203; crossRuth(32)=0.0283; crossRuth(33)=0.0300;
      crossRuth(34)=0.0373; crossRuth(35)=0.0471; crossRuth(36)=0.0685;
      crossRuth(37)=0.0914; crossRuth(38)=0.115; crossRuth(39)=0.138;
      crossRuth(40)=0.162; crossRuth(41)=0.184; crossRuth(42)=0.228;
      crossRuth(43)=0.267; crossRuth(44)=0.304; crossRuth(45)=0.337;
      crossRuth(46)=0.375; crossRuth(47)=0.409; crossRuth(48)=0.439;
      crossRuth(49)=0.466; crossRuth(50)=0.491; crossRuth(51)=0.513;
      crossRuth(52)=0.533; crossRuth(53)=0.551; crossRuth(54)=0.611;
      crossRuth(55)=0.689; crossRuth(56)=0.752; crossRuth(57)=0.804;
    }


   if(_ma==9)
    {
      _Z=1.; _A=1.; _density=1.; _radlength=1.;
      _R=1.;
    }
}




///////////////////////////////////////////////////////////////////////////
//
// NAME
//
//  Collimator::dumpCollimatorHits(const Integer &herdNo, Ostream &sout)
//
// DESCRIPTION
//  Calculate and dump histogram of number of particles lost to
//  each aperture node.
//
// PARAMETERS
//    herdNo - Reference to a macro-particle herd
//    sout  - Reference to a stream to dump the info.
//
// RETURNS
//     Nothing
//
///////////////////////////////////////////////////////////////////////////

Void Collimator::dumpCollimatorHits(const Integer &herdNo, Ostream &sout)
{

  CollimatorBase *tlp;
  Node *n1;
  MacroPart *mp;
  String wname;
  Real totPerc=0.;

  if ( (herdNo & All_Mask) > Particles::nHerds)
    except(badHerdNo);
   
  SyncPart *sp = SyncPart::safeCast(syncP(0));
  mp = MacroPart::safeCast(mParts((herdNo & All_Mask)-1)); 

  Integer nLostMacros = mp->_lostOnes._nLostMacros;
  Integer nLostMacrosGlobal = nLostMacros;  
  CollimatorBase::_syncLostOrHits(nLostMacros,nLostMacrosGlobal);

  sout.setf(ios::left, ios::adjustfield);

  if (nCollimators == 0)
    {
      if(Collimator::nMPIrank == 0) sout << "\n No Collimators \n";
      return;
    }
  if(Collimator::nMPIrank == 0) sout<<setw(10)<<"Name"<<setw(6)<<"Index"<<setw(16);
  if(Collimator::nMPIrank == 0) sout<<"Position [m]"<<setw(12)<<"# Hits";
  if(Collimator::nMPIrank == 0) sout<<setw(12)<<"# Absorbed"<<setw(15)<<"Efficiency";
  if(Collimator::nMPIrank == 0) sout<<setw(11)<<"% of Total\n";
  for (Integer i=1; i<= nCollimators; i++)
     {
        tlp = CollimatorBase::safeCast(CollimatorPointers(i-1));
	tlp->nameOut(wname);
        CollimatorBase::_syncLostOrHits(tlp->_nHits,tlp->_nHitsGlobal);
        CollimatorBase::_syncLostOrHits(tlp->_nLost,tlp->_nLostGlobal);
	if (wname.length() > 10) wname = wname.before(9) + "*";
	if(Collimator::nMPIrank == 0) sout<<setw(11) << wname <<setw(0);
	if(Collimator::nMPIrank == 0) sout<< setw(5) << tlp->_oindex;
	if(Collimator::nMPIrank == 0) sout<< setw(15) << tlp->_position; 
	if(Collimator::nMPIrank == 0) sout<< setw(12) << tlp->_nHitsGlobal;     
	if(Collimator::nMPIrank == 0) sout<< setw(12) << tlp->_nLostGlobal;
	if(tlp->_nLostGlobal == 0 || tlp->_nHitsGlobal == 0 )
          {
	    if(Collimator::nMPIrank == 0) sout<< setw(15) <<"0";
	  }
	else
	  {
	    if(Collimator::nMPIrank == 0) sout<< setw(15) << float(tlp->_nLostGlobal)/
	      float(tlp->_nHitsGlobal);
	  }
	if( nLostMacrosGlobal == 0)
	  {
	     if(Collimator::nMPIrank == 0) sout<< setw(15) <<"0\n";
	  }
	else
	  {
	    totPerc+=float(tlp->_nLostGlobal)/
	      float(nLostMacrosGlobal)*100;
	    if(Collimator::nMPIrank == 0) sout<< setw(15) << (float(tlp->_nLostGlobal)/
	      float(nLostMacrosGlobal))*100<<"\n";
	  }
     }
  if(Collimator::nMPIrank == 0) sout<<"\n Percent lost to collimators: "<<totPerc<<"\n";
}




float fstep(float s, float r_o, float pr_o, float theta)
{
  float a, b, c, d;
  float f_x;
  float n_o = 3.;

  a=Sqr(n_o*theta)/3.; 
  b=Sqr(pr_o); 
  c=2*r_o*pr_o; 
  d=Sqr(r_o);
  f_x=a*s*s*s - b*s*s - c*s - d;

  return f_x;
}


float rfunc(float x, float p, float fac1)
{
  return sin(x)*exp(-fac1*pow(2*p*sin(x/2), 2))/pow(sin(x/2),4);
}


int zbrak(float (*fx)(float, float, float, float), float x1, float x2, 
	  int n, float xb1[], float xb2[], int &nb, float param1, 
	  float param2, float param3)
{
  int nbb, i;
  float x, fp, fc, dx;
  
  nbb=0;
  dx=(x2-x1)/n;
  fp=(*fx)(x=x1, param1, param2, param3);
  for (i=1; i<=n; i++) {
    fc=(*fx)(x += dx, param1, param2, param3);
    if(fc*fp <= 0.0) {
      xb1[++nbb]=x-dx;
      xb2[nbb]=x;
      if(nb == nbb) return 0;
    }
    fp=fc;
  }
  nb=nbb;
}



float rtbis(float (*func)(float, float, float, float), float x1, float x2, 
	    float xacc, float param1, float param2, float param3)
{
  int j;
  float dx, f, fmid, xmid, rtb;
  
  f=(*func)(x1, param1, param2, param3);
  fmid=(*func)(x2, param1, param2, param3);
  if(f*fmid >= 0.0) cout<<"Root must be bracketed for bisection in rtbis\n";
  rtb = f < 0.0 ? (dx=x2-x1, x1) : (dx = x1-x2, x2);
  for (j=1; j <= JMAX; j++) {
    fmid = (*func)(xmid=rtb+(dx *= 0.5), param1, param2, param3);
    if (fmid <= 0.0) rtb = xmid; 
    if (fabs(dx) < xacc || fmid == 0.0) return rtb;
  }
  cout<<"Too many bisections in rtbis \n";
  return 0.0;
}



float bessj0(float x)
{
	float ax,z;
	double xx,y,ans,ans1,ans2;

	if ((ax=fabs(x)) < 8.0) {
		y=x*x;
		ans1=57568490574.0+y*(-13362590354.0+y*(651619640.7
			+y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
		ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
			+y*(59272.64853+y*(267.8532712+y*1.0))));
		ans=ans1/ans2;
	} else {
		z=8.0/ax;
		y=z*z;
		xx=ax-0.785398164;
		ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
			+y*(-0.2073370639e-5+y*0.2093887211e-6)));
		ans2 = -0.1562499995e-1+y*(0.1430488765e-3
			+y*(-0.6911147651e-5+y*(0.7621095161e-6
			-y*0.934935152e-7)));
		ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
	}
	return ans;
}



float bessj1(float x)
{
	float ax,z;
	double xx,y,ans,ans1,ans2;

	if ((ax=fabs(x)) < 8.0) {
		y=x*x;
		ans1=x*(72362614232.0+y*(-7895059235.0+y*(242396853.1
			+y*(-2972611.439+y*(15704.48260+y*(-30.16036606))))));
		ans2=144725228442.0+y*(2300535178.0+y*(18583304.74
			+y*(99447.43394+y*(376.9991397+y*1.0))));
		ans=ans1/ans2;
	} else {
		z=8.0/ax;
		y=z*z;
		xx=ax-2.356194491;
		ans1=1.0+y*(0.183105e-2+y*(-0.3516396496e-4
			+y*(0.2457520174e-5+y*(-0.240337019e-6))));
		ans2=0.04687499995+y*(-0.2002690873e-3
			+y*(0.8449199096e-5+y*(-0.88228987e-6
			+y*0.105787412e-6)));
		ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
		if (x < 0.0) ans = -ans;
	}
	return ans;
}

#undef EPS
#undef JMAX
#define EPS 1.0e-6
#define JMAX 20

float qsimp(float (*func)(float, float, float), float a, float b, 
float p, float fac1)
{
	float trapzd(float (*func)(float, float, float), float a, 
		     float b, int n, float p, float fac1);
	void nrerror(char error_text[]);
	int j;
	float s,st,ost,os;

	ost = os = -1.0e30;
	for (j=1;j<=JMAX;j++) {
		st=trapzd(func,a,b,j,p,fac1);
		s=(4.0*st-ost)/3.0;
		if (j > 5)
			if (fabs(s-os) < EPS*fabs(os) ||
				(s == 0.0 && os == 0.0)) return s;
		os=s;
		ost=st;
	}
	//nrerror("Too many steps in routine qsimp");
	return 0.0;
}
#undef EPS
#undef JMAX


/* note #undef's at end of file */
#define FUNC(x, p, fac1) ((*rfunc)(x, p, fac1))

float trapzd(float (*rfunc)(float, float, float), float a, 
	     float b, int n, float p, float fac1)
{
	float x,tnm,sum,del;
	static float s;
	int it,j;

	if (n == 1) {
		return (s=0.5*(b-a)*(FUNC(a, p, fac1)+FUNC(b, p, fac1)));
	} else {
		for (it=1,j=1;j<n-1;j++) it <<= 1;
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x, p, fac1);
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}
#undef FUNC
